package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.util.TreeNode;
import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.SearchService.ExtractionDepth;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.FileTime;
import java.text.ParseException;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;

@Service
public class FileService implements DisposableBean {

    private final Logger log = LoggerFactory.getLogger(FileService.class);

    private final Map<Path, ScheduledFuture<?>> futures = new ConcurrentHashMap<>();

    private static final String ZIP = "zip";

    @Autowired
    private ZipRepackagingService zipRepackagingService;

    @Autowired
    protected TokenProvider tokenProvider;

    @Autowired
    protected MetaDataRepository metaDataRepository;

    @Autowired
    protected GitLabRepository gitLabRepository;

    @Autowired
    protected GitlabService gitlabService;

    @Autowired
    protected PluginManagementService pluginManagementService;

    @Autowired
    protected ShoppingBasketService shoppingBasketService;

    @SuppressWarnings("PMD.DoNotUseThreads")
    private final ScheduledExecutorService executor = Executors.newScheduledThreadPool(Runtime.getRuntime().availableProcessors());

    @Override
    public void destroy() {
        futures.values().forEach(future -> future.cancel(true));
        futures.clear();
    }

    /**
     * Schedule the deletion of the given path with a given delay
     *
     * @param path           The path that should be deleted
     * @param delayInMinutes The delay in minutes after which the path should be deleted
     */
    public void scheduleForDeletion(Path path, long delayInMinutes) {
        log.debug("Removing [{}] scheduled for deletion", path);
        ScheduledFuture<?> future = executor.schedule(
            () -> {
                try {
                    if (Files.exists(path)) {
                        if (Files.isDirectory(path)) {
                            log.info("Delete directory {}", path);
                            FileUtils.deleteDirectory(path.toFile());
                        } else {
                            log.info("Delete file {}", path);
                            Files.delete(path);
                        }
                    }
                    futures.remove(path);
                } catch (IOException e) {
                    log.error("Deleting the file {} did not work: {}", path, e.getMessage());
                }
            },
            delayInMinutes,
            TimeUnit.MINUTES
        );

        futures.put(path, future);
    }

    /**
     * Method used to check if a directory is empty or not
     * https://www.baeldung.com/java-check-empty-directory
     *
     * @param path to the directory to check
     * @return true if directory is empty, false otherwise
     * @throws IOException
     */
    public boolean isEmpty(Path path) throws IOException {
        if (Files.isDirectory(path)) {
            try (Stream<Path> entries = Files.list(path)) {
                return entries.findFirst().isEmpty();
            }
        }

        return false;
    }

    public File copyInputStreamToFile(ZipInputStream inputStream, File file) {
        // append = false
        try (ZipOutputStream outputStream = new ZipOutputStream(new FileOutputStream(file, false))) {
            copyZip(inputStream, outputStream, null);
            return file;
        } catch (IOException e) {
            log.error("Error while copying input stream to file", e);
            return null;
        }
    }

    /**
     * copies a zipped input stream into the existing zipped output stream.
     * @param zis
     * @param zos
     * @param parent
     * @throws IOException
     */
    public void copyZip(ZipInputStream zis, ZipOutputStream zos, String parent) throws IOException {
        for (ZipEntry ze = zis.getNextEntry(); ze != null; ze = zis.getNextEntry()) {
            String newName = StringUtils.isEmpty(parent) ? ze.getName() : parent + "/" + ze.getName();
            ZipEntry newZE = new ZipEntry(newName);
            newZE.setComment(ze.getComment());
            FileTime creationTime = ze.getCreationTime();
            if (creationTime != null) {
                newZE.setCreationTime(creationTime);
            }

            zos.putNextEntry(newZE);
            byte[] buffer = new byte[1024];
            for (int read = zis.read(buffer); read != -1; read = zis.read(buffer)) {
                zos.write(buffer, 0, read);
            }
            zos.closeEntry();
        }

        zis.close();
    }

    public ZipInputStream getZippedExercise(
        ExerciseId exerciseId,
        Optional<User> user,
        ExtractionDepth recursive,
        Set<ExerciseId> alreadyTreatedExerciseIds
    ) throws IOException {
        final TreeNode<SearchResultDTO> exerciseTree = metaDataRepository.getExerciseCollectionTree(exerciseId.getRootExerciseId(), user);
        @SuppressWarnings("PMD.CloseResource")
        final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
        // first we try with user credentials!

        Optional<TreeNode<SearchResultDTO>> filter = Optional.of(exerciseTree);
        try (
            InputStream checkAuthorization = gitLabApi
                .getRepositoryApi()
                .getRepositoryArchive(Long.toString(exerciseTree.getData().getProject().getProject_id()), "HEAD", ZIP);
        ) {
            // if the previous gitlab requests succeeds, the user has full access to the repository.
            // no filtering is required
            if (checkAuthorization != null) {
                filter = Optional.empty(); // return everything
            }
        } catch (GitLabApiException e) {
            // if the previous gitlab throws an exception, we expect, that this is because of permission problems.
            // we check for public visibility, and construct the filter from the except array.
            log.warn("no access?: {}", e.getMessage());
        }

        try (
            InputStream repositoryArchive = gitLabRepository
                .getAdminGitLabApi()
                .getRepositoryApi()
                .getRepositoryArchive(Long.toString(exerciseTree.getData().getProject().getProject_id()), "HEAD", ZIP)
        ) {
            if (exerciseTree.getData().getMetadata().getPublicVisibility() == null && filter.isPresent()) {
                return null;
            }

            InputStream project = zipRepackagingService.rePackageGitLabProjectZip(
                new ZipInputStream(
                    gitLabRepository
                        .getAdminGitLabApi()
                        .getRepositoryApi()
                        .getRepositoryArchive(Long.toString(exerciseTree.getData().getProject().getProject_id()), "HEAD", ZIP)
                ),
                "from project " + exerciseId.getProjectId(),
                filter,
                exerciseId.getPath()
            );

            if (recursive == ExtractionDepth.JUST_PROJECT) {
                return new ZipInputStream(project);
            } else {
                alreadyTreatedExerciseIds.add(exerciseId); // Side effects are always dangerous :-(
                return extendZipByChildren(exerciseId, new ZipInputStream(project), user, recursive, alreadyTreatedExerciseIds);
            }
        } catch (GitLabApiException exception) {
            log.error(exception.getMessage());
            return null;
        }
    }

    /**
     *
     * @param exerciseId the exercise id
     * @param existingZip a zip input stream that should be amended by children, may be null
     * @param user the user to check for permissions
     * @param recursive the recursion type
     * @param alreadyTreatedExerciseIds list of already treated Exercise Ids (in order to avoid endless recursions)
     * @return an zipped output stream containing also children
     * @throws IOException
     */
    private ZipInputStream extendZipByChildren(
        ExerciseId exerciseId,
        ZipInputStream existingZip,
        Optional<User> user,
        ExtractionDepth recursive,
        Set<ExerciseId> alreadyTreatedExerciseIds
    ) throws IOException {
        if (recursive == ExtractionDepth.JUST_PROJECT) {
            return existingZip;
        }

        SearchResultDTO exerciseMD = metaDataRepository.getExerciseById(exerciseId, user);
        ExtractionDepth childRecursion = (recursive == ExtractionDepth.WITH_CHILDREN)
            ? ExtractionDepth.JUST_PROJECT
            : ExtractionDepth.WITH_DESCENDANTS;
        if (exerciseMD != null && exerciseMD.getFile().getChildren() != null) {
            File file = File.createTempFile("exercise" + exerciseId.getProjectId() + "recursive", ZIP);

            try (ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(file))) {
                if (existingZip != null) {
                    copyZip(existingZip, zos, null);
                }

                for (String child : exerciseMD.getFile().getChildren()) {
                    handleChild(child, user, childRecursion, alreadyTreatedExerciseIds, zos, exerciseId);
                }
            } catch (IOException e) {
                log.error("Error while creating temp file for recursive export", e);
            }

            return new ZipInputStream(new FileInputStream(file));
        }
        // nothing copied
        return existingZip;
    }

    private void handleChild(
        String child,
        Optional<User> user,
        ExtractionDepth childRecursion,
        Set<ExerciseId> alreadyTreatedExerciseIds,
        ZipOutputStream zos,
        ExerciseId parentExerciseId
    ) throws IOException {
        ExerciseId childExerciseId = null;
        try {
            childExerciseId = ExerciseId.fromString(child);
        } catch (ParseException e) {
            log.error("Error while parsing child id {}", child);
        }
        if (childExerciseId == null) {
            return;
        }

        if (childExerciseId.isDescendantOf(parentExerciseId)) {
            try (ZipInputStream zis = extendZipByChildren(childExerciseId, null, user, childRecursion, alreadyTreatedExerciseIds)) {
                copyZip(zis, zos, child);
            } catch (IOException e) {
                log.error("Cannot extract children {} {}", child, e.getMessage());
            }
        } else {
            try (ZipInputStream zis = getZippedExercise(childExerciseId, user, childRecursion, alreadyTreatedExerciseIds)) {
                copyZip(zis, zos, child);
            } catch (IOException e) {
                log.error("Cannot extract children {} {}", child, e.getMessage());
            }
        }
    }
}
