import { ComponentFixture, TestBed } from '@angular/core/testing';
import { HttpResponse } from '@angular/common/http';
import { HttpClientTestingModule } from '@angular/common/http/testing';
import { FormBuilder } from '@angular/forms';
import { ActivatedRoute } from '@angular/router';
import { RouterTestingModule } from '@angular/router/testing';
import { of, Subject, from } from 'rxjs';

import { SavedSearchesFormService } from './saved-searches-form.service';
import { SavedSearchesService } from '../service/saved-searches.service';
import { ISavedSearches } from '../saved-searches.model';

import { IUser } from 'app/entities/user/user.model';
import { UserService } from 'app/entities/user/user.service';

import { SavedSearchesUpdateComponent } from './saved-searches-update.component';

describe('SavedSearches Management Update Component', () => {
  let comp: SavedSearchesUpdateComponent;
  let fixture: ComponentFixture<SavedSearchesUpdateComponent>;
  let activatedRoute: ActivatedRoute;
  let savedSearchesFormService: SavedSearchesFormService;
  let savedSearchesService: SavedSearchesService;
  let userService: UserService;

  beforeEach(() => {
    TestBed.configureTestingModule({
      imports: [HttpClientTestingModule, RouterTestingModule.withRoutes([])],
      declarations: [SavedSearchesUpdateComponent],
      providers: [
        FormBuilder,
        {
          provide: ActivatedRoute,
          useValue: {
            params: from([{}]),
          },
        },
      ],
    })
      .overrideTemplate(SavedSearchesUpdateComponent, '')
      .compileComponents();

    fixture = TestBed.createComponent(SavedSearchesUpdateComponent);
    activatedRoute = TestBed.inject(ActivatedRoute);
    savedSearchesFormService = TestBed.inject(SavedSearchesFormService);
    savedSearchesService = TestBed.inject(SavedSearchesService);
    userService = TestBed.inject(UserService);

    comp = fixture.componentInstance;
  });

  describe('ngOnInit', () => {
    it('Should call User query and add missing value', () => {
      const savedSearches: ISavedSearches = { id: 456 };
      const user: IUser = { id: 56993 };
      savedSearches.user = user;

      const userCollection: IUser[] = [{ id: 65208 }];
      jest.spyOn(userService, 'query').mockReturnValue(of(new HttpResponse({ body: userCollection })));
      const additionalUsers = [user];
      const expectedCollection: IUser[] = [...additionalUsers, ...userCollection];
      jest.spyOn(userService, 'addUserToCollectionIfMissing').mockReturnValue(expectedCollection);

      activatedRoute.data = of({ savedSearches });
      comp.ngOnInit();

      expect(userService.query).toHaveBeenCalled();
      expect(userService.addUserToCollectionIfMissing).toHaveBeenCalledWith(
        userCollection,
        ...additionalUsers.map(expect.objectContaining)
      );
      expect(comp.usersSharedCollection).toEqual(expectedCollection);
    });

    it('Should update editForm', () => {
      const savedSearches: ISavedSearches = { id: 456 };
      const user: IUser = { id: 76974 };
      savedSearches.user = user;

      activatedRoute.data = of({ savedSearches });
      comp.ngOnInit();

      expect(comp.usersSharedCollection).toContain(user);
      expect(comp.savedSearches).toEqual(savedSearches);
    });
  });

  describe('save', () => {
    it('Should call update service on save for existing entity', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<ISavedSearches>>();
      const savedSearches = { id: 123 };
      jest.spyOn(savedSearchesFormService, 'getSavedSearches').mockReturnValue(savedSearches);
      jest.spyOn(savedSearchesService, 'update').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ savedSearches });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.next(new HttpResponse({ body: savedSearches }));
      saveSubject.complete();

      // THEN
      expect(savedSearchesFormService.getSavedSearches).toHaveBeenCalled();
      expect(comp.previousState).toHaveBeenCalled();
      expect(savedSearchesService.update).toHaveBeenCalledWith(expect.objectContaining(savedSearches));
      expect(comp.isSaving).toEqual(false);
    });

    it('Should call create service on save for new entity', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<ISavedSearches>>();
      const savedSearches = { id: 123 };
      jest.spyOn(savedSearchesFormService, 'getSavedSearches').mockReturnValue({ id: null });
      jest.spyOn(savedSearchesService, 'create').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ savedSearches: null });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.next(new HttpResponse({ body: savedSearches }));
      saveSubject.complete();

      // THEN
      expect(savedSearchesFormService.getSavedSearches).toHaveBeenCalled();
      expect(savedSearchesService.create).toHaveBeenCalled();
      expect(comp.isSaving).toEqual(false);
      expect(comp.previousState).toHaveBeenCalled();
    });

    it('Should set isSaving to false on error', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<ISavedSearches>>();
      const savedSearches = { id: 123 };
      jest.spyOn(savedSearchesService, 'update').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ savedSearches });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.error('This is an error!');

      // THEN
      expect(savedSearchesService.update).toHaveBeenCalled();
      expect(comp.isSaving).toEqual(false);
      expect(comp.previousState).not.toHaveBeenCalled();
    });
  });

  describe('Compare relationships', () => {
    describe('compareUser', () => {
      it('Should forward to userService', () => {
        const entity = { id: 123 };
        const entity2 = { id: 456 };
        jest.spyOn(userService, 'compareUser');
        comp.compareUser(entity, entity2);
        expect(userService.compareUser).toHaveBeenCalledWith(entity, entity2);
      });
    });
  });
});
