package at.ac.uibk.gitsearch.web.rest;

import static org.hamcrest.CoreMatchers.hasItem;
import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.Matchers.empty;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.config.ElasticsearchConfiguration;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService.ValidationResultDTO;
import co.elastic.clients.elasticsearch.ElasticsearchClient;
import co.elastic.clients.json.jackson.JacksonJsonpMapper;
import co.elastic.clients.transport.ElasticsearchTransport;
import co.elastic.clients.transport.rest_client.RestClientTransport;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.json.JsonMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Stream;
import org.apache.http.HttpHost;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO;
import org.elasticsearch.client.RestClient;
import org.junit.Ignore;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.ApplicationContext;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

@SpringBootTest(classes = GitsearchApp.class)
@ExtendWith(MockitoExtension.class)
@AutoConfigureMockMvc
class ValidationCheckerResourceIT {

    @Autowired
    private MockMvc validationCheckerMockMvc;

    @Autowired
    private ApplicationContext applicationContext;

    @Test
    @Timeout(value = 10, unit = TimeUnit.MINUTES)
    @Transactional
    void testCorrect() throws Exception {
        UserProvidedMetadataDTO metaData = createBasicMetaData();
        metaData.setKeyword(new String[] { "Variables" });
        metaData.setLicense("CC0 4.0");
        metaData.setLanguage(new String[] { "en" });
        metaData.setDifficulty("medium");

        validationCheckerMockMvc
            .perform(
                post("/api/validateMetaData")
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(metaData))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isOk())
            .andExpect(jsonPath("$.errors").value(is(empty())))
            .andExpect(jsonPath("$.warnings").value(is(empty())));
    }

    @Test
    @Transactional
    void testMissingMandatory() throws Exception {
        UserProvidedMetadataDTO metaData = new UserProvidedMetadataDTO();

        validationCheckerMockMvc
            .perform(
                post("/api/validateMetaData")
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(metaData))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isOk())
            .andExpect(jsonPath("$.errors").value(hasItem("The field keyword is mandatory")))
            .andExpect(jsonPath("$.warnings").value(is(empty())));
    }

    @Test
    @Transactional
    void testMissingMandatoryTopLevel() throws Exception {
        UserProvidedMetadataDTO metaData = new UserProvidedMetadataDTO();

        validationCheckerMockMvc
            .perform(
                post("/api/validateMetaData?topLevel=true")
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(metaData))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isOk())
            .andExpect(jsonPath("$.errors").value(hasItem("The field keyword is mandatory")))
            .andExpect(jsonPath("$.warnings").value(is(empty())));

        validationCheckerMockMvc
            .perform(
                post("/api/validateMetaData?topLevel=false")
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(metaData))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isOk())
            .andExpect(jsonPath("$.errors").value(hasItem("The field keyword is mandatory")))
            .andExpect(jsonPath("$.warnings").value(is(empty())));
    }

    @Test
    @Timeout(value = 10, unit = TimeUnit.MINUTES)
    @Transactional
    void testUnknownEntry() throws Exception {
        UserProvidedMetadataDTO metaData = createBasicMetaData();
        metaData.setKeyword(new String[] { "Variables", "VariablenX" });
        metaData.setLicense("CC0 4.0");
        metaData.setLanguage(new String[] { "de" });
        // Get all the likesList
        validationCheckerMockMvc
            .perform(
                post("/api/validateMetaData")
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(metaData))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isOk())
            .andDo(MockMvcResultHandlers.print())
            .andExpect(jsonPath("$.errors").value(is(empty())))
            .andExpect(
                jsonPath("$.warnings")
                    .value(
                        hasItem(
                            is(
                                "The field keyword contains an unknown entry 'VariablenX'. Did you mean 'Variablen' instead of 'VariablenX'?"
                            )
                        )
                    )
            );
    }

    private UserProvidedMetadataDTO createBasicMetaData() {
        UserProvidedMetadataDTO metaData = new UserProvidedMetadataDTO();
        final UserProvidedMetadataDTO.Person person1 = new UserProvidedMetadataDTO.Person();
        person1.setName("Toni Tester");
        person1.setEmail("Toni.Tester@codeability.org");
        person1.setAffiliation("Codeability Org");
        metaData.setCreator(new UserProvidedMetadataDTO.Person[] { person1 });
        metaData.setPublisher(new UserProvidedMetadataDTO.Person[] { person1 });
        metaData.setFormat(new String[] { "pdf" });
        return metaData;
    }

    // @Test // for some reason maven does not ignore tests
    @Timeout(value = 1, unit = TimeUnit.HOURS)
    @Ignore("for manual testing only: Use it to analyze an arbitrary gitsearch metadata repository dev/prod/local")
    @Transactional
    void testAllMetaDataEntries() throws Exception {
        ElasticsearchClient elasticsearchAPIClient = getOtherRestHighLevelClient();

        MetaDataRepository otherMDRepository = new MetaDataRepository(elasticsearchAPIClient);
        applicationContext.getAutowireCapableBeanFactory().autowireBean(otherMDRepository);
        AtomicInteger entryCount = new AtomicInteger(0);

        List<String> validationResults = new ArrayList<>();

        final Stream<SearchResultDTO> entries = otherMDRepository.getAllResultsAsStream();

        entries.forEach(entry -> {
            UserProvidedMetadataDTO metaData = entry.getMetadata();

            // Get all the likesList
            MvcResult result;
            try {
                result =
                    validationCheckerMockMvc
                        .perform(
                            post("/api/validateMetaData")
                                .contentType(MediaType.APPLICATION_JSON)
                                .content(TestUtil.convertObjectToJsonBytes(metaData))
                                .with(csrf().asHeader())
                        )
                        .andExpect(status().isOk()) // .andDo(MockMvcResultHandlers.print())
                        // .andExpect(jsonPath("$.errors").value(hasItem("The field keyword is
                        // mandatory")))
                        // .andExpect(jsonPath("$.warnings").value(is(empty())))
                        .andReturn();
                ObjectMapper mapper = JsonMapper
                    .builder()
                    .configure(DeserializationFeature.READ_ENUMS_USING_TO_STRING, true)
                    .configure(MapperFeature.ACCEPT_CASE_INSENSITIVE_ENUMS, true)
                    .configure(DeserializationFeature.ACCEPT_SINGLE_VALUE_AS_ARRAY, true)
                    .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, true)
                    .build();
                ValidationResultDTO validationResult = mapper.readValue(
                    result.getResponse().getContentAsString(),
                    ValidationResultDTO.class
                );

                if (!validationResult.getErrors().isEmpty() || !validationResult.getWarnings().isEmpty()) {
                    StringBuilder message = new StringBuilder();
                    message.append(entryCount + " " + entry.getExerciseId() + ": " + metaData.getTitle() + "\n");
                    validationResult.getErrors().forEach(m -> message.append("  Error: " + m + "\n"));
                    validationResult.getWarnings().forEach(m -> message.append("   Warn: " + m + "\n"));
                    validationResults.add(message.toString());
                }
                entryCount.incrementAndGet();
            } catch (IOException e) {
                e.printStackTrace();
            } catch (Exception e) {
                e.printStackTrace();
            }
        });
        validationResults.forEach(mess -> System.out.println(mess));
    }

    public ElasticsearchClient getOtherRestHighLevelClient() throws MalformedURLException {
        final JacksonJsonpMapper mapper = new JacksonJsonpMapper();
        final JavaTimeModule timeModule = new JavaTimeModule();
        timeModule.addDeserializer(LocalDateTime.class, new ElasticsearchConfiguration.MoreRobustLocalDateTimeConverter());

        mapper.objectMapper().registerModule(timeModule);
        mapper.objectMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        ElasticsearchTransport transport = new RestClientTransport(buildClient(), mapper);

        // And create the API client
        return new ElasticsearchClient(transport);
    }

    private RestClient buildClient() throws MalformedURLException {
        String elasticSearchUrls = "http://localhost:49200";
        Assert.hasText(elasticSearchUrls, "[Assertion Failed] At least one host must be set.");

        ArrayList<HttpHost> httpHosts = new ArrayList<>();
        for (String host : elasticSearchUrls.split(",")) {
            URL hostUrl = new URL(host);
            httpHosts.add(new HttpHost(hostUrl.getHost(), hostUrl.getPort(), hostUrl.getProtocol()));
        }
        return RestClient.builder(httpHosts.toArray(new HttpHost[0])).build();
    }
}
